"""Expriment with setting pyproject versions to latest lock file versions."""

from pathlib import Path

from corallium.log import LOGGER
from corallium.tomllib import tomllib


def _collect_pyproject_versions(pyproject_text: str) -> dict[str, str]:
    """Return pyproject versions without version specification for possible replacement.

    Documentation: https://python-poetry.org/docs/dependency-specification

    """
    pyproject = tomllib.loads(pyproject_text)

    pyproject_versions: dict[str, str] = {}
    # for section in
    pyproject_groups = pyproject['tool']['poetry'].get('group', {})
    groups = [group.get('dependencies', []) for group in pyproject_groups.values()]
    for deps in [pyproject['tool']['poetry']['dependencies'], *groups]:
        for name, value in deps.items():
            if name == 'python':
                continue
            version = value if isinstance(value, str) else value.get('version')
            if not version or any(_c in version for _c in ',*!@/'):
                LOGGER.text('WARNING: requires manually review', name=name, version=version)
            else:
                pyproject_versions[name] = version.lstrip('~^<>=')
    return pyproject_versions


def _replace_pyproject_versions(
    lock_versions: dict[str, str],
    pyproject_versions: dict[str, str],
    pyproject_text: str,
) -> str:
    """Return pyproject text with replaced versions."""
    new_lines: list[str] = []
    active_section = ''
    for line in pyproject_text.split('\n'):
        if line.startswith('['):
            active_section = line
        elif '=' in line and 'dependencies' in active_section:
            name = line.split('=')[0].strip()
            if (lock_version := lock_versions.get(name)) and (pyproject_version := pyproject_versions.get(name)):
                versions = {'name': name, 'new_version': lock_version, 'old_version': pyproject_version}
                # TODO: Handle ">=3.0.0,<4"
                if pyproject_version != lock_version:
                    if pyproject_version in line:
                        new_lines.append(line.replace(pyproject_version, lock_version, 1))
                        LOGGER.text('Upgrade minimum package version', **versions)  # type: ignore[arg-type]
                        continue
                    LOGGER.warning(
                        'Could not set new version. Please do so manually and submit a bug report',
                        line=line,
                        **versions,
                    )
            elif lock_version and not pyproject_versions.get(name):
                LOGGER.text('WARNING: consider manually updating the version', new_version=lock_version)

        new_lines.append(line)
    return '\n'.join(new_lines)


def replace_versions(path_lock: Path) -> None:
    """Read packages from poetry.lock and update the versions in pyproject.toml.

    Args:
        path_lock: Path to the poetry.lock file

    Raises:
        NotImplementedError: if a lock file other that the poetry lock file is used

    """
    if path_lock.name != 'poetry.lock':
        msg = f'Expected a path to a "poetry.lock" file. Instead, received: "{path_lock.name}"'
        raise NotImplementedError(msg)

    lock = tomllib.loads(path_lock.read_text(encoding='utf-8', errors='ignore'))
    lock_versions = {dependency['name']: dependency['version'] for dependency in lock['package']}

    path_pyproject = path_lock.parent / 'pyproject.toml'
    pyproject_text = path_pyproject.read_text(encoding='utf-8')
    pyproject_versions = _collect_pyproject_versions(pyproject_text)

    path_pyproject.write_text(_replace_pyproject_versions(lock_versions, pyproject_versions, pyproject_text))



"""Lint CLI."""

from contextlib import suppress

from beartype.typing import Optional
from corallium.file_helpers import read_package_name
from invoke.context import Context

from calcipy.cli import task
from calcipy.invoke_helpers import run

from .executable_utils import PRE_COMMIT_MESSAGE, check_installed, python_dir, python_m

# ==============================================================================
# Linting


def _inner_task(
    ctx: Context,
    *,
    command: str,
    cli_args: str = '',
    run_as_module: bool = True,
    target: Optional[str] = None,
) -> None:
    """Shared task logic."""
    file_args = []
    with suppress(AttributeError):
        file_args = ctx.config.gto.file_args
    if file_args:
        target = ' '.join([f'"{_a}"' for _a in file_args])
    elif target is None:
        target = f'"./{read_package_name()}" ./tests'

    cmd = f'{python_m()} {command}' if run_as_module else f'{python_dir()}/{command}'
    run(ctx, f'{cmd} {target} {cli_args}'.strip())


@task(default=True)
def check(ctx: Context) -> None:
    """Run ruff as check-only."""
    _inner_task(ctx, command='ruff check')


@task(
    help={
        'unsafe': 'if provided, attempt even fixes considered unsafe',
    },
)
def fix(ctx: Context, *, unsafe: bool = False) -> None:
    """Run ruff and apply fixes."""
    cli_args = '--fix'
    if unsafe:
        cli_args += ' --unsafe-fixes'
    _inner_task(ctx, command='ruff check', cli_args=cli_args)


@task()
def watch(ctx: Context) -> None:
    """Run ruff as check-only."""
    _inner_task(ctx, command='ruff check', cli_args='--watch')


# ==============================================================================
# Pre-Commit

ALL_PRE_COMMIT_HOOK_STAGES = [
    'commit',
    'merge-commit',
    'push',
    'prepare-commit-msg',
    'commit-msg',
    'post-checkout',
    'post-commit',
    'post-merge',
    'post-rewrite',
    'manual',
]


@task(
    help={
        'no_update': 'Skip updating the pre-commit hooks',
    },
)
def pre_commit(ctx: Context, *, no_update: bool = False) -> None:
    """Run pre-commit."""
    check_installed(ctx, executable='pre-commit', message=PRE_COMMIT_MESSAGE)

    run(ctx, 'pre-commit install')
    if not no_update:
        run(ctx, 'pre-commit autoupdate')

    stages_cli = ' '.join(f'--hook-stage {stg}' for stg in ALL_PRE_COMMIT_HOOK_STAGES)
    run(ctx, f'pre-commit run --all-files {stages_cli}')



"""Check for stale packages."""

from __future__ import annotations

import asyncio
import json
import time
from collections.abc import Awaitable
from dataclasses import asdict, dataclass, field
from datetime import timezone
from functools import partial
from itertools import starmap
from pathlib import Path
from typing import Any

import arrow
import httpx
from arrow import Arrow
from beartype.typing import Callable, TypeVar
from corallium.file_helpers import LOCK
from corallium.log import LOGGER
from corallium.tomllib import tomllib

from calcipy import can_skip  # Required for mocking can_skip.can_skip

CALCIPY_CACHE = Path('.calcipy_packaging.lock')
"""Path to the packaging lock file."""


def arrow_as_string_factory(data: list[tuple[str, Any]]) -> dict[str, Any]:
    """`asdict.dict_factory` to serialize Arrow types.

    Note: if `dict_factory` composition is needed, use functools.reduce and see:
    <https://cucumbersome.hashnode.dev/combine-multiple-dict-factories-in-asdict>

    """
    return {key: str(value) if isinstance(value, Arrow) else value for key, value in data}


@dataclass
class _HostedPythonPackage:
    """Representative information for a python package hosted on some domain."""

    # Note: "releases" was removed from the versioned URL: https://warehouse.pypa.io/api-reference/json.html#release
    name: str
    version: str
    domain: str = field(default='https://pypi.org/pypi/{name}/json')
    datetime: Arrow | None = field(default=None)
    latest_version: str = field(default='')
    latest_datetime: Arrow | None = field(default=None)

    @classmethod
    def from_data(cls, **kwargs) -> _HostedPythonPackage:
        for date_key in ('datetime', 'latest_datetime'):
            if value := kwargs.get(date_key):
                kwargs[date_key] = arrow.get(value)
        return cls(**kwargs)

    def model_dump_json(self) -> str:
        data = asdict(self, dict_factory=arrow_as_string_factory)
        return json.dumps(data)


async def _get_release_date(package: _HostedPythonPackage) -> _HostedPythonPackage:
    """Retrieve release date metadata for the specified package.

    Args:
        package: `_HostedPythonPackage`

    Returns:
        _HostedPythonPackage: updated with release date metadata from API

    """
    # Retrieve the JSON summary for the specified package
    json_url = package.domain.format(name=package.name)
    async with httpx.AsyncClient(timeout=10) as client:
        res = await client.get(json_url, timeout=30)  # nosem
        res.raise_for_status()
        res_json = res.json()
        if not (releases := res_json['releases']):
            msg = f'Failed to locate "releases" or "urls" from {json_url}: {res_json}'
            raise RuntimeError(msg)

    # Also retrieve the latest release date of the package looking through all releases
    release_dates = {
        arrow.get(release_data[0]['upload_time_iso_8601']): version
        for version, release_data in releases.items()
        if release_data
    }
    inverse_release_dates = {_v: key for key, _v in release_dates.items()}
    try:
        package.datetime = inverse_release_dates[package.version]
    except KeyError:  # pragma: no cover
        msg = f'Could not locate {package} in {res_json}. Please wait and try again later'
        raise RuntimeError(msg) from None
    package.latest_datetime = max([*release_dates])
    package.latest_version = release_dates[package.latest_datetime]
    return package


def _read_cache(path_pack_lock: Path = CALCIPY_CACHE) -> dict[str, _HostedPythonPackage]:
    """Read the cached packaging information.

    Args:
        path_pack_lock: Path to the lock file. Default is `CALCIPY_CACHE`

    Returns:
        Dict[str, _HostedPythonPackage]: the cached packages

    """
    if not path_pack_lock.is_file():
        path_pack_lock.write_text('{}', encoding='utf-8')
    old_cache: dict[str, dict[str, str]] = json.loads(path_pack_lock.read_text(encoding='utf-8'))
    return {package_name: _HostedPythonPackage.from_data(**meta_data) for package_name, meta_data in old_cache.items()}


_OpReturnT = TypeVar('_OpReturnT')


async def _rate_limited(
    operations: list[Callable[[], Awaitable[_OpReturnT]]],
    max_per_interval: int,
    interval_sec: int,
    max_delay: int | None = None,
) -> list[_OpReturnT]:
    """Naive semaphore-based rate limiting.

    For more performant and flexible limiting, see: <https://pypi.org/project/pyrate-limiter>

    Args:
        operations: list of no-argument callable functions to run
        max_per_interval: maximum number of concurrent operations per interval
        interval_sec: length of interval in seconds
        max_delay: maximum runtime before quietly stopping

    Returns:
        List[_OpReturnT]: list of return values from operations up to max_delay time, if set

    """
    if not operations:
        return []

    initial_start = time.monotonic()
    sem = asyncio.BoundedSemaphore(value=max_per_interval)

    async def task(idx: int, op: Callable[[], Awaitable[_OpReturnT]]) -> tuple[_OpReturnT | None, float]:
        """Return result. Rudimentary rate limiting by waiting to acquire the semaphore, then sleeping if necessary."""
        if max_delay and (time.monotonic() - initial_start) > max_delay:
            return (None, 0.0)
        max_idle = interval_sec if (count_input - idx) >= max_per_interval else 0
        async with sem:
            start = time.monotonic()
            result = await op()
            if (idle := max_idle - (time.monotonic() - start)) > 0:
                await asyncio.sleep(idle)
            return (result, idle)

    count_input = len(operations)
    tasks = await asyncio.gather(*starmap(task, enumerate(operations)))
    results = [out[0] for out in tasks if out[0]]
    total_idle = sum(max([out[1], 0]) for out in tasks)
    count_output = len(results)
    LOGGER.info(
        'Completed rate limited operations',
        count_input=count_input,
        count_output=count_output,
        avg_idle_time=round(total_idle / count_output, 2),
        total=round(time.monotonic() - initial_start, 2),
    )
    return results


async def _collect_release_dates(
    packages: list[_HostedPythonPackage],
    old_cache: dict[str, _HostedPythonPackage] | None = None,
) -> list[_HostedPythonPackage]:
    """Use the cache to retrieve only metadata that needs to be updated.

    Args:
        packages: list of `_HostedPythonPackage`
        old_cache: cache data to compare against to limit network requests

    Returns:
        List[_HostedPythonPackage]: packages with updated release dates

    """
    if old_cache is None:
        old_cache = {}

    updated_packages: list[_HostedPythonPackage] = []
    missing_packages: list[_HostedPythonPackage] = []
    for package in packages:
        cached_package = old_cache.get(package.name)
        cached_version = '' if cached_package is None else cached_package.version
        if package.version != cached_version:
            missing_packages.append(package)
        elif cached_package:
            updated_packages.append(cached_package)

    async def fetch(package: _HostedPythonPackage) -> _HostedPythonPackage | None:
        try:
            return await _get_release_date(package)
        except httpx.HTTPError as exc:
            LOGGER.warning('Could not lock package', package=package, error=str(exc))
        return None

    updated_packages.extend(
        [
            result
            for result in await _rate_limited(
                [partial(fetch, pkg) for pkg in missing_packages],
                max_per_interval=3,
                interval_sec=3,
                max_delay=600,
            )
            if result
        ],
    )
    return updated_packages


def _write_cache(updated_packages: list[_HostedPythonPackage], path_pack_lock: Path = CALCIPY_CACHE) -> None:
    """Read the cached packaging information.

    Args:
        updated_packages: updated packages to store
        path_pack_lock: Path to the lock file. Default is `CALCIPY_CACHE`

    """
    new_cache = {pack.name: json.loads(pack.model_dump_json()) for pack in updated_packages}
    pretty_json = json.dumps(new_cache, indent=4, separators=(',', ': '), sort_keys=True)
    path_pack_lock.write_text(pretty_json + '\n', encoding='utf-8')


def _read_packages(path_lock: Path) -> list[_HostedPythonPackage]:
    """Read packages from lock file. Currently only support `poetry.lock`, but could support more in the future.

    Args:
        path_lock: Path to the lock file to parse

    Returns:
        List[_HostedPythonPackage]: packages found in the lock file

    Raises:
        NotImplementedError: if a lock file other that the poetry lock file is used

    """
    if path_lock.name != LOCK.name:
        msg = f'"{path_lock.name}" is not a currently supported lock type. Try "{LOCK.name}" instead'
        raise NotImplementedError(msg)

    lock = tomllib.loads(path_lock.read_text(errors='ignore'))
    return [
        _HostedPythonPackage.from_data(
            name=dependency['name'],
            version=dependency['version'],
        )
        for dependency in lock['package']
    ]


def _packages_are_stale(packages: list[_HostedPythonPackage], *, stale_months: int) -> bool:
    """Check for stale packages. Raise error and log all stale versions found.

    Args:
        packages: List of packages
        stale_months: cutoff in months for when a package might be stale enough to be a risk

    """

    def format_package(pack: _HostedPythonPackage) -> str:
        delta = pack.datetime.humanize()  # type: ignore[union-attr]
        latest = '' if pack.version == pack.latest_version else f' (*New version available: {pack.latest_version}*)'
        return f'{delta}: {pack.name} {pack.version}{latest}'

    stale_cutoff = arrow.now(timezone.utc).shift(months=-1 * stale_months)
    sorted_packages = sorted(packages, key=lambda x: x.datetime or stale_cutoff, reverse=True)
    stale_packages = [pack for pack in sorted_packages if not pack.datetime or pack.datetime < stale_cutoff]
    if stale_packages:
        LOGGER.warning(
            f'Found stale packages older than {stale_months} months',
            stale_packages=[format_package(_p) for _p in stale_packages],
        )
    LOGGER.text('Oldest packages:\n\t' + '\n\t'.join([format_package(_p) for _p in sorted_packages[-5:]]))
    return bool(stale_packages)


def check_for_stale_packages(*, stale_months: int, path_lock: Path = LOCK, path_cache: Path = CALCIPY_CACHE) -> bool:
    """Check for stale packages by reading from the cache, and updating if necessary.

    Args:
        stale_months: cutoff in months for when a package might be stale enough to be a risk
        path_lock: path to poetry lock file
        path_cache: path to calcipy package cache file

    """
    cached_packages = _read_cache(path_cache)
    if cached_packages and can_skip.can_skip(prerequisites=[path_lock], targets=[path_cache]):
        packages = [*cached_packages.values()]
    else:
        packages = asyncio.run(_collect_release_dates(_read_packages(path_lock), cached_packages))
        _write_cache(packages, path_cache)
    return _packages_are_stale(packages, stale_months=stale_months)



"""Create a namespace with all tasks that can be imported."""

from contextlib import suppress

from calcipy.collection import Collection

from .defaults import new_collection

# "ns" will be recognized by Collection.from_module(all_tasks)
# https://docs.pyinvoke.org/en/stable/api/collection.html#invoke.collection.Collection.from_module
ns = new_collection()
with suppress(RuntimeError):
    from . import cl

    ns.add_collection(Collection.from_module(cl))
with suppress(RuntimeError):
    from . import doc

    ns.add_collection(Collection.from_module(doc))
with suppress(RuntimeError):
    from . import lint

    ns.add_collection(Collection.from_module(lint))
with suppress(RuntimeError):
    from . import nox

    ns.add_collection(Collection.from_module(nox))
with suppress(RuntimeError):
    from . import pack

    ns.add_collection(Collection.from_module(pack))
with suppress(RuntimeError):
    from . import stale

    ns.add_collection(Collection.from_module(stale))
with suppress(RuntimeError):
    from . import tags

    ns.add_collection(Collection.from_module(tags))
with suppress(RuntimeError):
    from . import test

    ns.add_collection(Collection.from_module(test))
with suppress(RuntimeError):
    from . import types

    ns.add_collection(Collection.from_module(types))

__all__ = ('ns',)



"""Markdown table formatting."""

from __future__ import annotations

from collections.abc import Iterable
from itertools import starmap
from typing import Any


def format_table(
    headers: list[str],
    records: list[dict[str, Any]],
    delimiters: list[str] | None = None,
) -> str:
    """Returns a formatted Github Markdown table.

    Args:
        headers: ordered keys to use as column title
        records: list of key:row-value dictionaries
        delimiters: optional list to allow for alignment

    """
    table = [[str(_r[col]) for col in headers] for _r in records]
    widths = [max(len(row[col_idx].strip()) for row in [headers, *table]) for col_idx in range(len(headers))]

    def pad(values: list[str]) -> list[str]:
        return [val.strip().ljust(widths[col_idx]) for col_idx, val in enumerate(values)]

    def join(row: Iterable[str], spacer: str = ' ') -> str:
        return f'|{spacer}' + f'{spacer}|{spacer}'.join(row) + f'{spacer}|'

    def expand_delimiters(delim: str, width: int) -> str:
        expanded = '-' * (width + 2)
        if delim.startswith(':'):
            expanded = ':' + expanded[1:]
        if delim.endswith(':'):
            expanded = expanded[:-1] + ':'
        return expanded

    if delimiters:
        errors = []
        if len(delimiters) != len(headers):
            errors.append(f'Incorrect number of delimiters provided ({len(delimiters)}). Expected: ({len(headers)})')
        allowed_delimiters = {'-', ':-', '-:', ':-:'}
        if not all(delim in allowed_delimiters for delim in delimiters):
            errors.append(f'Delimiters must one of ({len(allowed_delimiters)}). Received: ({len(delimiters)})')
        if errors:
            raise ValueError(' and '.join(errors))

    lines = [
        join(pad(headers)),
        join(starmap(expand_delimiters, zip(delimiters or ['-'] * len(headers), widths)), ''),
        *[join(pad(row)) for row in table],
    ]
    return '\n'.join(lines)



# noqa: A005
"""Types CLI."""

from invoke.context import Context

from calcipy.cli import task
from calcipy.invoke_helpers import run

from .executable_utils import PYRIGHT_MESSAGE, check_installed, python_dir


@task()
def pyright(ctx: Context) -> None:
    """Run pyright using the config in `pyproject.toml`."""
    check_installed(ctx, executable='pyright', message=PYRIGHT_MESSAGE)
    run(ctx, 'pyright')


@task()
def mypy(ctx: Context) -> None:
    """Run mypy."""
    run(ctx, f'{python_dir()}/mypy')



"""Experiment with bumping the git tag using `griffe`."""

import griffe
import semver
from corallium.log import LOGGER
from griffe import BuiltinModuleError


def bump_tag(*, pkg_name: str, tag: str, tag_prefix: str) -> str:
    """Return either minor or patch change based on `griffe`.

    Note: major versions must be bumped manually

    """
    previous = griffe.load_git(pkg_name, ref=tag)
    current = griffe.load(pkg_name)

    breakages = [*griffe.find_breaking_changes(previous, current)]
    for breakage in breakages:
        try:
            LOGGER.text(breakage._explain_oneline())  # noqa: SLF001
        except BuiltinModuleError:  # noqa: PERF203
            LOGGER.warning(str(breakage))
        except Exception:
            LOGGER.exception(str(breakage))

    try:
        ver = semver.Version.parse(tag.replace(tag_prefix, ''))
    except ValueError:
        LOGGER.exception('Failed to parse tag', tag=tag)
        return ''
    new_ver = ver.bump_minor() if any(breakages) else ver.bump_patch()
    return f'{tag_prefix}{new_ver}'



"""Document CLI."""

import webbrowser
from contextlib import suppress
from pathlib import Path

from corallium.file_helpers import (
    MKDOCS_CONFIG,
    open_in_browser,
    read_yaml_file,
)
from invoke.context import Context
from invoke.exceptions import UnexpectedExit

from calcipy.cli import task
from calcipy.invoke_helpers import get_project_path, run
from calcipy.md_writer import write_template_formatted_md_sections

from .executable_utils import python_dir


def get_out_dir() -> Path:
    """Returns the mkdocs-specified site directory."""
    mkdocs_config = read_yaml_file(get_project_path() / MKDOCS_CONFIG)
    return Path(mkdocs_config.get('site_dir', 'releases/site'))


@task()
def build(ctx: Context) -> None:
    """Build documentation with mkdocs."""
    write_template_formatted_md_sections()
    run(ctx, f'{python_dir()}/mkdocs build --site-dir {get_out_dir()}')


def _is_mkdocs_local() -> bool:
    """Check if mkdocs is configured for local output.

    See notes on local-link configuration here: https://github.com/timothycrosley/portray/issues/65

    Additional information on using local search here: https://github.com/wilhelmer/mkdocs-localsearch

    Returns:
        bool: True if configured for local file output rather than hosted

    """
    mkdocs_config = read_yaml_file(get_project_path() / MKDOCS_CONFIG)
    return mkdocs_config.get('use_directory_urls') is False


@task()
def watch(ctx: Context) -> None:
    """Serve local documentation for local editing."""
    if _is_mkdocs_local():  # pragma: no cover
        path_doc_index = get_out_dir() / 'index.html'
        open_in_browser(path_doc_index)
    else:  # pragma: no cover
        webbrowser.open('http://localhost:8000')
        run(ctx, f'{python_dir()}/mkdocs serve --dirtyreload')


@task()
def deploy(ctx: Context) -> None:
    """Deploy docs to the Github `gh-pages` branch."""
    if _is_mkdocs_local():  # pragma: no cover
        raise NotImplementedError('Not yet configured to deploy documentation without "use_directory_urls"')

    with suppress(UnexpectedExit):
        run(ctx, 'pre-commit uninstall')  # To prevent pre-commit failures when mkdocs calls push
    run(ctx, f'{python_dir()}/mkdocs gh-deploy --force')
    with suppress(UnexpectedExit):
        run(ctx, 'pre-commit install')  # Restore pre-commit
